!
! HLL approximate Riemann solver
!
  implicit none
  integer :: i,n,nt
  real*8, allocatable, dimension(:) :: rho,u,P,rhou,E,f1,f2,f3
  real*8 :: rho_atm,gamma,xmin,xmax,dx,t,tend,dt,dtmax,cf,x
  logical :: exit_do
!
  ! Read the number of grid points and allocate memory
  open(1,file='riemann_initial',status='old')
  read(1,*) n ! number of grid points
  close(1)
  write(*,*) ' '
  write(*,*) 'Number of grid point = ',n
  if (n .le. 2) then 
     write(*,*) 'Number of grid points must be greater than 2.'
     stop
  end if
  allocate (rho(n))
  allocate (u(n))
  allocate (P(n))
  allocate (rhou(n))
  allocate (E(n))
  allocate (f1(n))
  allocate (f2(n))
  allocate (f3(n))

  call read_init(rho,u,P,rhou,E,gamma,rho_atm,xmin,xmax,dx,cf,tend,n)

  ! Timestepping
  t = 0.d0
  nt = 0
  exit_do = .false.
  write(*,*) ' '
  do 
     nt = nt+1
     call courant(rho,u,P,gamma,dx,dtmax,n)  ! Calculate max(dt) based on Courant condition
     dt = cf * dtmax
     if (t + dt .ge. tend) then 
        dt = tend - t
        exit_do = .true.
     end if
     write(*,*) 'Iteration #: ',nt,' to time t = ',t+dt
     call hll_flux(rho,u,P,rhou,E,gamma,f1,f2,f3,n)
     call update_vars(rho,rhou,E,f1,f2,f3,dt,dx,n)
     call primitives_and_bcs(rho,rhou,E,u,P,gamma,rho_atm,dx,xmin,n)
     t = t+dt
     if (exit_do) exit
  end do

  ! Output data at t=tend
  open(1,file='riemann_final',status='unknown')
  write(1,*) '# x     rho     u     P     at t = ',tend
  do i=1,n
     x = xmin + (i-0.5d0)*dx
     write(1,'(5ES20.10)') x,rho(i),u(i),P(i),P(i)/(gamma-1.d0)/rho(i)
  end do
  close(1)

  deallocate(rho,u,P,rhou,E,f1,f2,f3)
  stop
  end

! Read initial data
   subroutine read_init(rho,u,P,rhou,E,gamma,rho_atm,xmin,xmax,dx,cf,tend,n)
   implicit none
   integer :: n,i
   real*8, dimension(n) :: rho,u,P,rhou,E
   real*8 :: rhoL,rhoR,uL,uR,PL,PR,gamma,rho_atm
   real*8 :: rhouL, rhouR,EL,ER
   real*8 :: xmin,xmax,x0,dx,x,tend,cf
!
   open(1,file='riemann_initial',status='old')
   read(1,*)   
   read(1,*) gamma  ! Adiabatic index
   read(1,*) xmin 
   read(1,*) xmax
   read(1,*) x0     ! location of discontuity
   read(1,*) tend   ! Final time
   read(1,*) cf     ! Courant factor
   read(1,*) rhoL 
   read(1,*) uL
   read(1,*) PL
   read(1,*) rhoR
   read(1,*) uR
   read(1,*) PR
   close(1)

   write(*,*) ' '
   write(*,*) 'Adiabatic index gamma = ',gamma
   write(*,*) 'xmin, xmax: ',xmin,xmax
   write(*,*) 'Location of discontinuity  x0 = ',x0
   write(*,*) 'Final time tend = ',tend
   write(*,*) 'Courant factor cf = ',cf
   write(*,*) ' '
   write(*,*) 'Initial left state:'
   write(*,*) 'rhoL = ',rhoL
   write(*,*) 'uL = ',uL
   write(*,*) 'PL = ',PL
   write(*,*) ' ' 
   write(*,*) 'Initial right state:'
   write(*,*) 'rhoR = ',rhoR
   write(*,*) 'uR = ',uR
   write(*,*) 'PR = ',PR
   write(*,*) ' ' 

   ! Sanity check
   if (tend .le. 0.d0) then 
      write(*,*) 'tend must be greater than 0'
      stop
   end if
   if (cf .le. 0.d0 .or. cf .ge. 1.d0) then 
      write(*,*) 'cf must lie between 0 and 1'
      stop
   end if
   if (gamma .le. 1.d0) then 
      write(*,*) 'gamma must be greater than 1'
      stop
   end if
   if (rhoL .le. 0.d0) then 
      write(*,*) 'rhoL must be greater than 0.'
      stop
   end if
   if (rhoR .le. 0.d0) then 
      write(*,*) 'rhoR must be greater than 0.'
      stop
   end if
   if (PL .le. 0.d0) then 
      write(*,*) 'PL must be greater than 0.'
      stop
   end if
   if (PR .le. 0.d0) then 
      write(*,*) 'PR must be greater than 0.'
      stop
   end if
   if (xmin .ge. xmax) then 
      write(*,*) 'xmin must be less than xmax'
      stop
   end if
   if (x0 .le. xmin) then 
      write(*,*) 'x0 must be greater than xmin'
      stop
   end if
   if (x0 .ge. xmax) then 
      write(*,*) 'x0 must be smaller than xmax'
      stop
   end if

   dx = (xmax-xmin)/dble(n)

   rho_atm = 1.d-12*min(rhoL,rhoR)  ! 'Atmosphere' density

   rhouL = rhoL*uL
   rhouR = rhoR*uR
   EL = 0.5d0*rhouL*uL + PL/(gamma - 1.d0)
   ER = 0.5d0*rhouR*uR + PR/(gamma - 1.d0)

   do i=1,n
      x = (i-0.5d0)*dx + xmin
      if (x .lt. x0) then 
	 rho(i) = rhoL
         u(i) = uL
	 P(i) = PL
	 rhou(i) = rhouL
	 E(i) = EL
      else
	 rho(i) = rhoR
	 u(i) = uR
	 P(i) = PR
	 rhou(i) = rhouR
	 E(i) = ER
      end if
   end do

   end subroutine read_init

! Calculate max(dt) based on Courant condition
   subroutine courant(rho,u,P,gamma,dx,dtmax,n)  
   implicit none
   integer :: n,i
   real*8, dimension(n) :: rho,u,P
   real*8 :: gamma,dx,dtmax,lambda_max,c
!
   c = sqrt(gamma*P(1)/rho(1))  ! sound speed
   lambda_max = max( abs(u(1)+c), abs(u(1)-c) ) ! maximum characteristic speed
   dtmax = dx/lambda_max
   do i=2,n
      c = sqrt(gamma*P(i)/rho(i))  ! sound speed
      lambda_max = max( abs(u(i)+c), abs(u(i)-c) ) ! maximum characteristic speed
      dtmax = min( dtmax, dx/lambda_max)
   end do
   end subroutine courant

! Compute the HLL flux
   subroutine hll_flux(rho,u,P,rhou,E,gamma,f1,f2,f3,n)
   implicit none
   integer :: n,i
   real*8, dimension(n) :: rho,u,P,rhou,E,f1,f2,f3
   real*8 :: f1L,f1R,f2L,f2R,f3L,f3R
   real*8 :: SR,SL,cL,cR, gamma
   real*8 :: rhoR,rhoL,rhouR,rhouL,ER,EL
!
   do i=1,n-1
      f1L = rho(i)*u(i)
      f1R = rho(i+1)*u(i+1)
      f2L = rho(i)*u(i)*u(i) + P(i)
      f2R = rho(i+1)*u(i+1)*u(i+1) + P(i+1)
      f3L = u(i)*(E(i)+P(i))
      f3R = u(i+1)*(E(i+1)+P(i+1))
      rhoL = rho(i)
      rhoR = rho(i+1)
      rhouL = rhou(i)
      rhouR = rhou(i+1)
      EL = E(i)
      ER = E(i+1)
      cL = sqrt(gamma*P(i)/rho(i))
      cR = sqrt(gamma*P(i+1)/rho(i+1))
      SR = max( u(i)+cL, u(i+1)+cR)
      SL = min( u(i)-cL, u(i+1)-cR)
      if (SL .ge. 0.d0) then 
         f1(i) = f1L
	 f2(i) = f2L
	 f3(i) = f3L
      else if (SL .lt. 0.d0 .and. SR .gt. 0.d0) then 
         f1(i) = (SR*f1L - SL*f1R + SL*SR*(rhoR-rhoL))/(SR-SL)
         f2(i) = (SR*f2L - SL*f2R + SL*SR*(rhouR-rhouL))/(SR-SL)
         f3(i) = (SR*f3L - SL*f3R + SL*SR*(ER-EL))/(SR-SL)
      else 
         f1(i) = f1R
	 f2(i) = f2R
	 f3(i) = f3R
      end if
!if (i==30 .or. i==31) then 
!   write(*,*) i,SR,SL
!   write(*,*) f1L,f1R,f1(i)
!end if
   end do
   f1(n) = f1(n-1)
   f2(n) = f2(n-1)
   f3(n) = f3(n-1)
   end subroutine hll_flux

! Update the evolution variables
   subroutine update_vars(rho,rhou,E,f1,f2,f3,dt,dx,n)
   implicit none
   integer :: i,n
   real*8, dimension(n) :: rho,rhou,E,f1,f2,f3
   real*8 :: dt,dx,dtodx
!
   dtodx = dt/dx
   do i=2,n-1
      rho(i) = rho(i) - dtodx*(f1(i)-f1(i-1))
      rhou(i) = rhou(i) - dtodx*(f2(i)-f2(i-1))
      E(i) = E(i) - dtodx*(f3(i)-f3(i-1))
   end do
   end subroutine update_vars

! Primitive recovery and BCs
   subroutine primitives_and_bcs(rho,rhou,E,u,P,gamma,rho_atm,dx,xmin,n)
   implicit none
   integer :: n,i
   real*8, dimension(n) :: rho,rhou,E,u,P
   real*8 :: rho_atm,gamma,rhoeps,dx,xmin,x
   logical :: error
!
   ! Primitive recovery
   error = .false.
   do i=2,n
      u(i) = rhou(i)/rho(i)
      rhoeps = E(i) - 0.5d0*rhou(i)*u(i)
      P(i) = (gamma-1.d0)*rhoeps
      if (rho(i) .lt. 0.d0) then 
         write(*,*) 'Negative density at point i = ',i
         x = xmin + dx*(i-0.5d0)
         write(*,*) 'x, rho = ',x,rho(i)
         error = .true.
      end if     
      if (P(i) .lt. 0.d0) then 
         write(*,*) 'Negative pressure at point i = ',i
         x = xmin + dx*(i-0.5d0)
         write(*,*) 'x, P = ',x,P(i)
         error = .true.
      end if     
   end do
   if (error) stop

   ! Apply copy BC
   rho(1) = rho(2)
   u(1) = u(2)
   P(1) = P(2)
   rhou(1) = rhou(2)
   E(1) = E(2)
   rho(n) = rho(n-1)
   u(n) = u(n-1)
   P(n) = P(n-1)
   rhou(n) = rhou(n-1)
   E(n) = E(n-1)
   
   end subroutine primitives_and_bcs
